from typing import Iterable, Union
from ...common.marshalling import Api, bind_to_api, TestApi
from ...common import File, Loc, FileLike, IDecider, Locator
from .interface import IBrainboxService
from .service import BrainBoxService
from ...controllers import IController, ControllerRegistry, ControllerApi
import os
import requests
from .server import BrainBoxServer, BrainBoxServiceSettings
from pathlib import Path
from .serverless_test import ServerlessTest
from ...job_processing import AlwaysOnPlanner
from functools import partial


@bind_to_api(BrainBoxService)
class BrainBoxApi(Api, IBrainboxService):
    def __init__(self,
                 address: str = '127.0.0.1:18090',
                 cache_folder: Path|None = None
                 ):
        super().__init__(address)
        self.cache_folder = cache_folder if cache_folder is not None else Loc.cache_folder

    @property
    def controller_api(self) -> ControllerApi:
        return ControllerApi(self.address)

    def _process_download_arguments(
            self,
            fname: str|File,
            custom_file_path: Path|None = None,
            replace: bool = False) -> tuple[bool, str, Path]:
        if isinstance(fname, File):
            fname = fname.name
        if custom_file_path is None:
            custom_file_path = self.cache_folder/fname
        os.makedirs(custom_file_path.parent, exist_ok=True)
        should_download = True
        if custom_file_path.is_file():
            if not replace:
                should_download = False
        return should_download, fname, custom_file_path

    def _download_and_cache(self, fname: str, path: Path) -> bytes:
        address = f'http://{self.address}/cache/download/{fname}'
        response = requests.get(address)
        if response.status_code != 200:
            raise ValueError(f"Couldn't get file content for {address}\n" + response.text)
        content = response.content

        with open(path, 'wb') as stream:
            stream.write(content)
        return content


    def open_file(self, fname: str|File) -> File:
        should_download, fname, path = self._process_download_arguments(fname)
        if should_download:
            content = self._download_and_cache(fname, path)
            return File(fname, content)
        else:
            return File.read(path)

    def download(self, fname: str|File, custom_file_path: Path|None = None, replace: bool = False) -> Path:
        should_download, fname, path = self._process_download_arguments(fname, custom_file_path, replace)
        if should_download:
            self._download_and_cache(fname, path)
        return path

    def upload(self, filename: str, data: FileLike.Type):
        with FileLike(data, None) as stream:
            reply = requests.post(
                f'http://{self.address}/cache/upload/',
                files=(
                    (filename, stream),
                )
            )
        if reply.status_code != 200:
            raise ValueError(reply.text)


    class Test(TestApi['BrainBoxApi']):
        def __init__(self,
                     services: Iterable[Union[IDecider, IController]]|None = None,
                     run_controllers_in_default_environment: bool = True,
                     always_on_planner: bool = False,
                     stop_containers_at_termination: bool = True,
                     keep_folder: bool = False,
                     port: int = 18090,
                     locator: Locator|None = None
                     ):
            self.test_folder = Loc.create_test_folder('brainbox_test_runs')
            self.keep_folder = keep_folder
            if locator is None:
                locator = Locator(self.test_folder.path)
            registry = ControllerRegistry.discover_or_create(services)
            if not run_controllers_in_default_environment:
                registry.locator = locator
            settings = BrainBoxServiceSettings(
                registry,
                locator = locator,
                debug_output=True,
                stop_controllers_at_termination=stop_containers_at_termination,
                port=port
            )
            if always_on_planner:
                settings.planner = AlwaysOnPlanner(AlwaysOnPlanner.Mode.FindThenStart)
            super().__init__(partial(BrainBoxApi, cache_folder=locator.cache_folder), BrainBoxServer(settings))

        def __enter__(self):
            self.test_folder.__enter__()
            self._api = super().__enter__()
            return self._api

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._api.shutdown()
            super().__exit__(exc_type, exc_val, exc_tb)
            self.test_folder.__exit__(exc_type, exc_val, exc_tb)


    ServerlessTest = ServerlessTest






from typing import *
from abc import ABC, abstractmethod
from pathlib import Path
from .controller import IController

class DownloadableModel(ABC):
    @classmethod
    def from_dict(cls, dictionary: dict) -> Any:
        return None

    @abstractmethod
    def get_local_location(self, controller: IController) -> Path:
        pass

    @classmethod
    @abstractmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        pass



class IModelDownloadingController:
    @abstractmethod
    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        pass


    def _ensure_type(self, model, index=None) -> DownloadableModel:
        _type = self.get_downloadable_model_type()
        if isinstance(model, _type):
            return model
        if isinstance(model, dict):
            value = _type.from_dict(model)
            if value is not None:
                return value
            try:
                return _type(**model)
            except:
                raise ValueError(f"Cannot resolve {_type} from arguments {model}")
        if index is None:
            index = ''
        else:
            index = f'(at index {index})'
        raise ValueError(f"Model {index} is expected to be dict or {_type}, but was neither")


    def _normalize_model_array(self, models: Any) -> List[DownloadableModel]:
        _type = self.get_downloadable_model_type()
        if isinstance(models, dict):
            return [self._ensure_type(models)]
        elif isinstance(models, _type):
            return [models]
        try:
            models = list(models)
        except:
            raise ValueError(f"Models has to be dict, {_type} or list of those two, but wasn't")
        result = []
        for index, model in enumerate(models):
            result.append(self._ensure_type(model, index))
        return result


    def download_models(self, models: Any):
        controller = cast(IController,self)
        models = self._normalize_model_array(models)
        models_to_download = []
        for model in models:
            controller.context.logger.log(f"Checking model {model}")
            location = model.get_local_location(self)
            if location.is_dir() or location.is_file():
                controller.context.logger.log("Model is already downloaded")
                continue
            controller.context.logger.log("Model needs to be downloaded")
            models_to_download.append(model)
        if len(models_to_download) == 0:
            controller.context.logger.log("All models are downloaded")
        else:
            controller.context.logger.log(f"Downloading {len(models_to_download)} models")
            _type = self.get_downloadable_model_type()
            _type.download(models_to_download, self)











from typing import *
import subprocess
from yo_fluq import FileIO
from ...framework import IDecider


class Ffmpeg(IDecider):
    def concat(self,
               files: Iterable[str],
               target_extension: str = ".wav"
               ):
        paths = [self.cache_folder/file for file in files]
        control_content_array = [f"file '{p}'" for p in paths]
        control_content = '\n'.join(control_content_array)
        temp_file = self.cache_folder/(self.current_job_id+'.txt')
        FileIO.write_text(control_content, temp_file)
        output_file = self.cache_folder/(self.current_job_id+target_extension)
        args = [
            'ffmpeg',
            '-f',
            'concat',
            '-safe',
            '0',
            '-i',
            str(temp_file),
            '-y',
            str(output_file)
        ]
        try:
            subprocess.check_output(args)
            return output_file.name
        except subprocess.CalledProcessError as err:
            raise ValueError(f'FFMpeg returned non-zero value. arguments are\n{" ".join(args)}. Output\n{err.output}')

    def make_video_from_audio_and_image(
            self,
            image: str,
            audio: str,
            target_extension: str = '.mp4'
    ):
        output_filename = self.current_job_id+target_extension
        subprocess.call([
            'ffmpeg',
            '-y',
            '-loop',
            '1',
            '-i',
            str(self.cache_folder/image),
            '-i',
            str(self.cache_folder/audio),
            '-c:v',
            'libx264',
            '-c:a',
            'mp3',
            '-shortest',
            str(self.cache_folder/output_filename),
            '-y'
        ])
        return output_filename

    def create_silence_audio(self, duration_in_seconds: int = 1, sample_rate: int = 22000, target_extension=".wav"):
        output_filename = self.current_job_id+target_extension
        subprocess.call([
            'ffmpeg',
            '-f',
            'lavfi',
            '-i',
            f'anullsrc=r={sample_rate}:cl=mono',
            '-t',
            str(duration_in_seconds),
            str(self.cache_folder/output_filename)
        ])
        return output_filename




import os
from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel,
    File
)
from ...common import VOICEOVER_TEXT, check_if_its_sound
from .settings import ZonosSettings
from pathlib import Path


class ZonosController(
    DockerWebServiceController[ZonosSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
            mount_resource_folders= {
                'cache': '/home/app/.cache',
                'speakers': '/speakers',
                'voices': '/voices'
            },
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return ZonosSettings()

    def create_api(self):
        from .api import Zonos
        return Zonos()

    def post_install(self):
        path = self.resource_folder('cache')
        if path.is_dir() and len(os.listdir(path))==2:
            print('Already installed')
            return
        self.run_with_configuration(self.get_service_run_configuration(None).as_service_worker('--install'))

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Zonos

        speaker_sample = File.read(Path(__file__).parent/'container/lina.mp3')
        api.execute(BrainBoxTask.call(Zonos).train('test_speaker', [speaker_sample]))

        yield (
            TestReport
            .last_call(api)
            .href('train')
        )

        result = api.execute(BrainBoxTask.call(Zonos).voiceover(VOICEOVER_TEXT, 'test_speaker'))
        check_if_its_sound(api.open_file(result).content, tc)
        yield (
            TestReport
            .last_call(api)
            .href('voiceover')
            .result_is_file(File.Kind.Audio)
        )



DOCKERFILE = f'''
FROM pytorch/pytorch:2.6.0-cuda12.4-cudnn9-devel

{SmallImageBuilder.APT_INSTALL('espeak-ng git ffmpeg')}

RUN pip install uv

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

WORKDIR /home/app

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

{SmallImageBuilder.GIT_CLONE_AND_RESET('https://github.com/Zyphra/Zonos','Zonos','c6f7704',install=True)}

COPY . /home/app/

ENTRYPOINT ["python3","/home/app/main.py"]
'''

DEPENDENCIES = '''
aiofiles==23.2.1
annotated-types==0.7.0
anyio==4.8.0
argon2-cffi==23.1.0
argon2-cffi-bindings==21.2.0
arrow==1.3.0
asttokens==3.0.0
async-lru==2.0.4
attrs==25.1.0
babel==2.17.0
beautifulsoup4==4.13.3
bleach==6.2.0
blinker==1.9.0
causal-conv1d==1.5.0.post8
certifi==2025.1.31
cffi==1.17.1
charset-normalizer==3.4.1
click==8.1.8
clldutils==3.21.0
colorama==0.4.6
colorlog==6.9.0
comm==0.2.2
csvw==3.5.1
debugpy==1.8.12
decorator==5.1.1
defusedxml==0.7.1
dlinfo==2.0.0
einops==0.8.1
exceptiongroup==1.2.2
executing==2.2.0
fastapi==0.115.8
fastjsonschema==2.21.1
ffmpy==0.5.0
filelock==3.17.0
flash_attn==2.7.4.post1
Flask==3.1.0
fqdn==1.5.1
fsspec==2025.2.0
gradio==5.16.0
gradio_client==1.7.0
h11==0.14.0
httpcore==1.0.7
httpx==0.28.1
huggingface-hub==0.28.1
idna==3.10
inflect==7.5.0
ipykernel==6.29.5
ipython==8.32.0
ipywidgets==8.1.5
isodate==0.7.2
isoduration==20.11.0
itsdangerous==2.2.0
jedi==0.19.2
Jinja2==3.1.5
joblib==1.4.2
json5==0.10.0
jsonpointer==3.0.0
jsonschema==4.23.0
jsonschema-specifications==2024.10.1
jupyter-events==0.12.0
jupyter-lsp==2.2.5
jupyter_client==8.6.3
jupyter_core==5.7.2
jupyter_server==2.15.0
jupyter_server_terminals==0.5.3
jupyterlab==4.3.5
jupyterlab_pygments==0.3.0
jupyterlab_server==2.27.3
jupyterlab_widgets==3.0.13
kanjize==1.6.0
language-tags==1.2.0
lxml==5.3.1
mamba-ssm==2.2.4
Markdown==3.7
markdown-it-py==3.0.0
MarkupSafe==2.1.5
matplotlib-inline==0.1.7
mdurl==0.1.2
mistune==3.1.1
more-itertools==10.6.0
mpmath==1.3.0
nbclient==0.10.2
nbconvert==7.16.6
nbformat==5.10.4
nest-asyncio==1.6.0
networkx==3.4.2
ninja==1.11.1.3
notebook==7.3.2
notebook_shim==0.2.4
numpy==2.2.3
nvidia-cublas-cu12==12.4.5.8
nvidia-cuda-cupti-cu12==12.4.127
nvidia-cuda-nvrtc-cu12==12.4.127
nvidia-cuda-runtime-cu12==12.4.127
nvidia-cudnn-cu12==9.1.0.70
nvidia-cufft-cu12==11.2.1.3
nvidia-curand-cu12==10.3.5.147
nvidia-cusolver-cu12==11.6.1.9
nvidia-cusparse-cu12==12.3.1.170
nvidia-cusparselt-cu12==0.6.2
nvidia-nccl-cu12==2.21.5
nvidia-nvjitlink-cu12==12.4.127
nvidia-nvtx-cu12==12.4.127
orjson==3.10.15
overrides==7.7.0
packaging==24.2
pandas==2.2.3
pandocfilters==1.5.1
parso==0.8.4
pexpect==4.9.0
phonemizer==3.3.0
pillow==11.1.0
platformdirs==4.3.6
prometheus_client==0.21.1
prompt_toolkit==3.0.50
psutil==7.0.0
ptyprocess==0.7.0
pure_eval==0.2.3
pycparser==2.22
pydantic==2.10.6
pydantic_core==2.27.2
pydub==0.25.1
Pygments==2.19.1
pylatexenc==2.10
pyparsing==3.2.1
python-dateutil==2.9.0.post0
python-json-logger==3.2.1
python-multipart==0.0.20
pytz==2025.1
PyYAML==6.0.2
pyzmq==26.2.1
rdflib==7.1.3
referencing==0.36.2
regex==2024.11.6
requests==2.32.3
rfc3339-validator==0.1.4
rfc3986==1.5.0
rfc3986-validator==0.1.1
rich==13.9.4
rpds-py==0.22.3
ruff==0.9.6
safehttpx==0.1.6
safetensors==0.5.2
segments==2.2.1
semantic-version==2.10.0
Send2Trash==1.8.3
shellingham==1.5.4
six==1.17.0
sniffio==1.3.1
soundfile==0.13.1
soupsieve==2.6
stack-data==0.6.3
starlette==0.45.3
SudachiDict-full==20250129
SudachiPy==0.6.10
sympy==1.13.1
tabulate==0.9.0
terminado==0.18.1
tinycss2==1.4.0
tokenizers==0.21.0
tomli==2.2.1
tomlkit==0.13.2
torch==2.6.0
torchaudio==2.6.0
tornado==6.4.2
tqdm==4.67.1
traitlets==5.14.3
transformers==4.48.3
triton==3.2.0
typeguard==4.4.1
typer==0.15.1
types-python-dateutil==2.9.0.20241206
typing_extensions==4.12.2
tzdata==2025.1
uri-template==1.3.0
uritemplate==4.1.1
urllib3==2.3.0
uv==0.5.31
uvicorn==0.34.0
wcwidth==0.2.13
webcolors==24.11.1
webencodings==0.5.1
websocket-client==1.8.0
websockets==14.2
Werkzeug==3.1.3
widgetsnbextension==4.0.13
'''




from brainbox.doc import run_all
from brainbox import BrainBox
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('-p', '--port', default='8090')
parser.add_argument('-o', '--host', default='127.0.0.1')

if __name__ == '__main__':
    args = parser.parse_args()
    api = BrainBox.Api(f'{args.host}:{args.port}')
    api.wait(1)
    run_all(api)


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, IModelDownloadingController, DownloadableModel
)
from .settings import OllamaSettings, OllamaModel
from pathlib import Path


class OllamaController(DockerWebServiceController[OllamaSettings], IModelDownloadingController):
    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return OllamaModel

    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            None
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is None:
            raise ValueError(f"`parameter` cannot be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 11434},
            mount_resource_folders={'main': "/root/.ollama"},
            mount_top_resource_folder=False,
            command_line_arguments=['run',parameter],
            run_as_root=True,
        )

    def get_default_settings(self):
        return OllamaSettings()

    def create_api(self):
        from .api import Ollama
        return Ollama()

    def post_install(self):
        self.download_models(self.settings.models_to_install)

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Ollama
        model = self.settings.models_to_install[0]
        prompt = "The recipe for the borsch is as follows:"
        api.execute(BrainBoxTask.call(Ollama, model.name).completions_json(prompt=prompt))
        yield TestReport.last_call(api).href('completions-json').with_comment("Returns json for completions with detailed reply")

        api.execute(BrainBoxTask.call(Ollama, model.name).completions(prompt=prompt))
        yield TestReport.last_call(api).href('completions-txt').with_comment("Returns only the text result")

        prompt = "Give me the recipe of the borsch."
        api.execute(BrainBoxTask.call(Ollama, model.name).question_json(prompt=prompt))
        yield TestReport.last_call(api).href('question-json').with_comment("Returns json for question mode with detailed reply")

        api.execute(BrainBoxTask.call(Ollama, model.name).question(prompt=prompt))
        yield TestReport.last_call(api).href('question-txt').with_comment("Returns only the text result")



DOCKERFILE = '''
FROM ollama/ollama

COPY run.sh /bin/run.sh

ENTRYPOINT ["/bin/bash", "/bin/run.sh"]
'''




import os

from ..app import BrainBoxApi, IBrainBoxTask
from typing import TypeVar, Generic, Any
from pathlib import Path
from dataclasses import dataclass
from yo_fluq import FileIO

T = TypeVar("T")

@dataclass
class BrainBoxCommandCacheItem:
    id: str|None
    result: Any = None


class BrainBoxCommandCache(Generic[T]):
    def __init__(self, file: Path):
        self.file = file
        if file.name.endswith('.json'):
            self._as_json = True
        elif file.name.endswith('.pkl') or file.name.endswith('.pickle'):
            self._as_json = False
        else:
            raise ValueError(f"Filename for cache must end with .json or .pkl/.pickle, but was: {file.name}")

    def read(self) -> BrainBoxCommandCacheItem:
        if not self.file.parent.is_dir():
            return BrainBoxCommandCacheItem(None, None)
        if not self.file.is_file():
            return BrainBoxCommandCacheItem(None, None)
        if self._as_json:
            return BrainBoxCommandCacheItem(**FileIO.read_json(self.file))
        else:
            return FileIO.read_pickle(self.file)
    def get_cached_result(self) -> T|None:
        return self.read().result

    def update(self, id: str, result: T|None = None):
        os.makedirs(self.file.parent, exist_ok=True)
        if self._as_json:
            FileIO.write_json(dict(id=id, result=result), self.file)
        else:
            FileIO.write_pickle(BrainBoxCommandCacheItem(id, result))


class BrainBoxCommand(Generic[T]):
    Cache = BrainBoxCommandCache

    def __init__(self,
                 task: IBrainBoxTask,
                 cache: BrainBoxCommandCache|None = None
                 ):
        self.task = task
        self.cache = cache

    def with_cache(self, file: Path) -> 'BrainBoxCommand':
        return BrainBoxCommand(self.task, BrainBoxCommandCache(Path(file)))

    def execute(self, api: BrainBoxApi) -> T:
        if self.cache is not None:
            cached = self.cache.read()
            if cached.result is not None:
                return cached.result
            elif cached.id is not None:
                return self._join(api, cached.id)
        self.add(api)
        return self.join(api)


    def add(self, api: BrainBoxApi):
        if self.cache is not None:
            cached = self.cache.read()
            if cached.id is not None:
                return
        api.add(self.task)
        if self.cache is not None:
            self.cache.update(self.task.get_resulting_id())

    def _join(self, api: BrainBoxApi, id: str) -> T:
        result = api.join(id)
        result = self.task.postprocess_result(result, api)
        if self.cache is not None:
            self.cache.update(id, result)
        return result

    def join(self, api: BrainBoxApi) -> T:
        return self._join(api, self.task.get_resulting_id())




import subprocess
from server import EspeakPhonemizerApp
import argparse
import sys

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-n', '--notebook', action='store_true')

    args = parser.parse_args()
    print(f"Running with arguments\n{args}")

    if args.notebook:
        subprocess.call([sys.executable, '-m', 'notebook', '--allow-root', '--port', '8899', '--ip', '0.0.0.0', "--NotebookApp.token=''"], cwd='/repo')
        exit(0)


    EspeakPhonemizerApp().create_app().run('0.0.0.0',8080)



from .command import Command
from .executor import IExecutor
from .local_executor import LocalExecutor
from .ssh_executor import SSHExecutor
from .file_system import IFileSystem
from .local_file_system import LocalFileSystem
from .ssh_file_system import SSHFileSystem
from .machine import Machine


from .api_utils import ApiUtils
from .decider import IDecider, DeciderContext, ISelfManagingDecider
from .file import File
from .file_like import FileLike
from .fork import Fork
from .loc import Loc, Locator
from .logger import Logger
from .signature_processor import SignatureProcessor
from .file_io import FileIO
from .HTML import HTML


from .app import *
from .task import *


class BrainBox:
    ITask = IBrainBoxTask
    Api = BrainBoxApi
    Task = BrainBoxTask
    ExtendedTask = BrainBoxExtendedTask
    CombinedTask = BrainBoxCombinedTask
    Command = BrainBoxCommand



import json

from ..interface import ControllerServiceStatus
from ....common import HTML



def create_status_page(status: ControllerServiceStatus):
    html = []
    if status.currently_installing_container is not None:
        html.append(f'<a href="/html/controllers/installation_report">Installing {status.currently_installing_container}</a>')
    html.append('''
    <table cellspacing="0" border="1">
    <tr>
    <td>Name</td>
    <td>Installed/Size</td>
    <td>Installation commands</td>
    <td>Instances</td>
    <td>Run</td>
    <td>Self-test report</td>
    </tr>
    ''')
    for row in status.containers:
        html.append('<tr>')
        html.append(f'<td>{row.name}</td>')
        html.append(f'<td>{"☑" if row.installation_status.installed else "☐"} {str(row.installation_status.size)}</td>')

        if row.installation_status.dockerless_controller:
            html.append('<td>&nbsp;</td>'*4+'</tr>')
            continue

        html.append(f'<td>')
        html.append('<table border="0"><tr>')
        html.append("<td>")
        html.append(HTML.button('/controllers/install', "Install", dict(decider=row.name, join=False)))
        html.append('</td><td>')
        html.append(HTML.button('/controllers/uninstall', "Uninstall", dict(decider=row.name), active=row.installation_status.installed))
        html.append('</td><td>')
        html.append(HTML.button('/controllers/uninstall', "Purge", dict(decider=row.name, purge=True), active=row.installation_status.installed))
        html.append('</td><td>')
        html.append(HTML.button('/controllers/self_test', "Self-test", dict(decider=row.name), active=row.installation_status.installed))
        html.append('</td>')
        html.append('</tr></table>')
        html.append('</td>')

        html.append('<td><table border="0">')
        for instance in row.instances:
            html.append("<tr>")
            html.append(f"<td>{instance.instance_id}</td>")
            html.append(f"<td>{instance.parameter}</td>")
            if instance.address is not None:
                html.append(f'<td><a href="http://{instance.address}" target="_blank">{instance.address}</a></td>')
            else:
                html.append('<td>&nbsp;</td>')
            html.append(f"<td>")
            html.append(HTML.button("/controllers/stop", "Stop", dict(decider=row.name, instance_id=instance.instance_id)))
            html.append(f"</td></tr>")
        html.append("</table></td>")

        html.append("<td>")
        parameter_name = 'parameter_'+row.name
        html.append(f'<input type="text" id="{parameter_name}"/>')
        html.append(HTML.button('/controllers/run', "Run", dict(decider=row.name), dict(parameter=parameter_name), active=row.installation_status.installed))
        html.append("</td>")

        html.append("<td>")
        if row.has_self_test_report:
            html.append(f"<a href=/html/controllers/self_test_report/{row.name}>Self-test report</a>")
            if row.has_errors_in_self_test_report:
                html.append(f'⚠')
            html.append(HTML.button('/controllers/delete_self_test', "Delete", dict(decider=row.name)))
        else:
            html.append("&nbsp;")
        html.append("</td>")

        html.append("</tr>")

    html.append('</table>')
    return HTML.page("Brainbox: Controllers", html)